#!/usr/bin/env python
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import logging
import string
import sys
import os
import time
import platform

import usb.core
import usb.util
import usb.backend.libusb1
import serial.tools.list_ports

# fix of broken paths for windows
if not sys.path:
    sys.path = []
path = os.getcwd()
if path not in sys.path:
    sys.path.append(path)

import config


class USBDetector:
    #NOTE Don't use try:... except (usb.core.USBError) for pyusb calls, because pyusb is unstable(up to v1.0.2)
    #and can produce all kinds of unpredictable exceptions

    @staticmethod
    def format_vid_or_pid(vid_or_pid): #cuts "0x", fill with zeroes if needed, doing case up
        if vid_or_pid:
            return hex(vid_or_pid)[2:].zfill(4).upper()

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.libusb_backend = usb.backend.libusb1.get_backend(find_library=self.get_libusb_path)
        self.forced_types = config.get_settings().get('forced_types', {})
        self.no_empty_serials = config.get_settings().get('no_empty_serials')
        self.backward_compatible_serials = config.get_settings().get("backward_compatible_serials")

    def calculate_bus_offset(self):
        bus_offset = 0
        if sys.platform.startswith('win'):
            bus_offset = 0
        return bus_offset            

    def get_libusb_path(self, lib):
        if sys.platform.startswith('win'):
            python_version = platform.architecture()[0]
            if '64' in python_version:
                libusb_name = 'libusb-1.0-64.dll'
            else:
                libusb_name = 'libusb-1.0.dll'
            pythons_folder = os.path.dirname(os.path.abspath(sys.executable))
            embeded_python_backend_path = os.path.join(pythons_folder, libusb_name)
            if os.path.isfile(embeded_python_backend_path):
                return embeded_python_backend_path
            local_backend_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), libusb_name)
            if os.path.isfile(local_backend_path):
                return local_backend_path

    def detect_devices(self):
        retries = 5
        while retries: # the retries is for mac, because it can suddenly refuse to detect usb devices
            try:
                devices = usb.core.find(find_all=True, backend=self.libusb_backend)
            except Exception as e:
                self.logger.exception("Exception in pyusb: %s" % e)
                devices = None
            if not devices:
                if retries == 5:
                    message = "Warning: no USB devices detected. This is an error. Detection retries left: %d" % retries
                    self.logger.warning(message)
                time.sleep(0.5)
                retries -= 1
            else:
                return devices
        else:
            return []

    def get_printers_list(self, ignore_profiles=False, add_descriptions=False):
        try:
            serial_ports = serial.tools.list_ports.comports()
        except Exception as e:
            self.logger.warning(e)
            return []
        unused_serial_ports = [x for x in serial_ports if x[2] != "n/a"]
        printers = []
        profiles = config.get_profiles()
        for dev in self.detect_devices():
            try:
                VID = USBDetector.format_vid_or_pid(dev.idVendor)
                PID = USBDetector.format_vid_or_pid(dev.idProduct)
            except:
                continue
            try:
                bus = dev.bus
                port_numbers = dev.port_numbers
                if port_numbers is None:
                    port_numbers = (dev.port_number,)
                PRT = "%d-%d" % (bus, port_numbers[0])
                PRT += ".%d" * len(port_numbers[1:]) % port_numbers[1:]
            except:  
                PRT = ""
            SNR = ""
            usb_info = {'VID': VID, 'PID': PID, 'SNR': SNR, 'PRT': PRT}
            forced_type = self.forced_types.get(VID + ":" + PID)
            if forced_type:
                usb_info['forced_type'] = forced_type
            profile = self.find_profile_by_vid_pid(profiles, VID, PID, forced_type)
            if profile or ignore_profiles: 
                # Warning! Crazy serial number can lead to selection of wrong printer profile 
                if not profile or not profile.get('crazy_serial_number'):
                    try:
                        SNR = usb.util.get_string(dev, dev.iSerialNumber)
                    except:
                        pass
                    if SNR:
                        for symbol in SNR:
                            if symbol not in string.printable:
                                SNR = ""
                                break
                        usb_info['SNR'] = SNR
                usb_info['COM'] = self.get_serial_port_name(unused_serial_ports, VID, PID, SNR, PRT)
                if add_descriptions:
                    try:
                        usb_info["manufacturer"] = str(dev.manufacturer)
                    except: #some device descriptors will make libusb to crash on this action
                        pass
                    try:
                        usb_info["product"] = str(dev.product)
                    except: #some device descriptors will make libusb to crash on this action
                        pass
                printers.append(usb_info)
        for printer in printers:
            if not printer.get('COM'):
                printer['COM'] = self.get_serial_port_name(unused_serial_ports, printer['VID'], printer['PID'])
            if not printer['SNR']:
                printer['SNR'] = self.get_snr_by_serial_port_name(printer['COM'], serial_ports)
            if not printer['SNR'] and self.no_empty_serials:
                printer['SNR'] = printer['PRT']
        self.modify_identical_printers_id(printers)
        return printers

    def get_snr_by_serial_port_name(self, serial_port_name, all_serial_ports):
        for port in all_serial_ports:
            if port.device == serial_port_name:
                return port.serial_number

    def get_serial_port_name(self, unused_serial_ports, vid, pid, snr="", prt=""):
        for port in unused_serial_ports:
            if vid == USBDetector.format_vid_or_pid(port.vid) and pid == USBDetector.format_vid_or_pid(port.pid):
                if prt and port.location and prt in port.location:
                    unused_serial_ports.remove(port)
                    return port.device
        for port in unused_serial_ports:
            if vid == USBDetector.format_vid_or_pid(port.vid) and pid == USBDetector.format_vid_or_pid(port.pid):
                if snr and snr.upper() == port.serial_number.upper():
                    unused_serial_ports.remove(port)
                    return port.device
        for port in unused_serial_ports:
            if vid == USBDetector.format_vid_or_pid(port.vid) and pid == USBDetector.format_vid_or_pid(port.pid):
                unused_serial_ports.remove(port)
                return port.device
        return ""

    def find_profile_by_vid_pid(self, profiles, vid, pid, forced_type=None):
        if profiles:
            for profile in profiles:
                if [vid, pid] in profile.get('vids_pids', []):
                    if not forced_type or profile.get('alias') == forced_type:
                        return profile

    # find printers with same VID, PID, SNR and include COM name in their SNRs
    def modify_identical_printers_id(self, printers):
        for checked_printer in printers:
            for printer in printers:
                if printer['VID'] == checked_printer['VID'] and printer['PID'] == checked_printer['PID'] and\
                   printer['SNR'] == checked_printer['SNR'] and\
                   (printer['COM'] != checked_printer['COM'] or printer['PRT'] != checked_printer['PRT']):
                      self.update_duplicate_printer_info(printer)
                      self.update_duplicate_printer_info(checked_printer)

    def update_duplicate_printer_info(self, printer):
        printer['ID_CLONE_WARNING'] = True
        if printer['COM'] and self.backward_compatible_serials:
            if not printer['SNR']:
                printer['SNR'] = ''
            printer['SNR'] = printer['SNR'] + printer['COM'].split("/")[-1]
        elif printer['PRT']:
            if not printer['SNR']:
                printer['SNR'] = ''
            printer['SNR'] = printer['SNR'] + printer['PRT']

if __name__ == '__main__':
    print('PyUSB v' + usb.__version__)
    print('PySerial v' + serial.__version__)
    detector = USBDetector()
    printers = detector.get_printers_list(ignore_profiles=True, add_descriptions=True)
    print("\nAll devices:")
    if printers:
        for printer in printers:
            print(printer)
    else:
        print(None)
    if config.get_profiles(): #we dont need this we we cant import config and profiles
        printers = detector.get_printers_list()
        print("\nPrinters with profiles:")
        if printers:
            for printer in printers:
                print(printer)
        else:
            print(None)
    printers = [x for x in printers if x.get('COM')]
    print("\nDevices with serial port:")
    if printers:
        for printer in printers:
            print(printer)
    else:
        print(None)
    print("\nCOM port devices by pyserial:")
    for port in serial.tools.list_ports.comports():
        print("Device:%s HWID:%s" % (port.device, port.hwid))
